class ClientHandler:
    """
    A class for handling clients. If the client is already created, return it. Otherwise, create it.
    """

    _clients = {}

    def __init__(self, client_initializer):
        """
        Initialize the ClientHandler.

        Args:
        - client_initializer: A function that initializes the client. E.g. `google.cloud.bigquery.Client`.
        """
        self.client_initializer = client_initializer
        self.module_name = client_initializer.__module__
        try:
            self.client_name = client_initializer.__name__
        except AttributeError:
            self.client_name = client_initializer.__class__.__name__
        self.initializer_name = f"{self.module_name}.{self.client_name}"

    def get(self, force_refresh=False, **kwargs):
        """
        Get the client. If the client is already created, return it. Otherwise, create it.

        Args:
        - force_refresh: If True, create a new client even if the client is already created.
        - kwargs: The arguments to pass to the client_initializer. E.g. `project="my-project"`, `location="europe-west2"`.

        Returns:
        - client: The client.
        """
        input_key = frozenset(kwargs.items())
        client_key = (self.initializer_name, input_key)
        if client_key in ClientHandler._clients and not force_refresh:
            client = ClientHandler._clients[client_key]
        else:
            client = self.client_initializer(**kwargs)
            ClientHandler._clients[client_key] = client
        return client


if __name__ == "__main__":
    from google.cloud import bigquery, firestore
    import time

    time_now = time.time()
    bq_client1 = ClientHandler(bigquery.Client).get()
    time_after_bq_client1 = time.time()
    print(f"Time taken for bq_client1: {time_after_bq_client1 - time_now}")
    bq_client2 = ClientHandler(bigquery.Client).get()
    time_after_bq_client2 = time.time()
    print(f"Time taken for bq_client2: {time_after_bq_client2 - time_after_bq_client1}")
    bq_client3 = ClientHandler(bigquery.Client).get(force_refresh=True)
    time_after_bq_client3 = time.time()
    print(f"Time taken for bq_client3: {time_after_bq_client3 - time_after_bq_client2}")

    fs_client1 = ClientHandler(firestore.Client).get()
    time_after_fs_client1 = time.time()
    print(f"Time taken for fs_client1: {time_after_fs_client1 - time_after_bq_client3}")
    fs_client2 = ClientHandler(firestore.Client).get(project="my-project")
    time_after_fs_client2 = time.time()
    print(f"Time taken for fs_client2: {time_after_fs_client2 - time_after_fs_client1}")
    fs_client3 = ClientHandler(firestore.Client).get(force_refresh=True)
    time_after_fs_client3 = time.time()
    print(f"Time taken for fs_client3: {time_after_fs_client3 - time_after_fs_client2}")
    fs_client4 = ClientHandler(firestore.Client).get(project="my-project")
    time_after_fs_client4 = time.time()
    print(f"Time taken for fs_client4: {time_after_fs_client4 - time_after_fs_client3}")

    assert bq_client1 is bq_client2
    assert bq_client1 is not bq_client3

    assert fs_client1 is not fs_client2
    assert fs_client1 is not fs_client3
    assert fs_client2 is fs_client4

    assert len(ClientHandler._clients) == 3



import os
import json
import logging
import importlib
import collections.abc


LIST_LIKE_TYPES = (list, tuple, set, frozenset, collections.abc.KeysView)


def try_import(module_name, origin_module=None, errors="raise"):
    """
    Attempt to dynamically import a module. If the import fails, raise an informative ImportError.

    Args:
    - module_name (str): The name of the module to be imported.
    - origin_module (str, optional): The module that is attempting to perform the import.
    """
    from gcp_pal.config import PYPI_NAMES

    try:
        return importlib.import_module(module_name)
    except (ImportError, ModuleNotFoundError) as e:
        if "is not a package" in str(e):
            return None
        elif "No module name" in str(e):
            try:
                # Try to import the module, e.g. "google.cloud.bigquery" and get the class, e.g. "Client"
                class_name = module_name.split(".")[-1]
                try_module_name = ".".join(module_name.split(".")[:-1])
                module = importlib.import_module(try_module_name)
                try:
                    return getattr(module, class_name)
                except AttributeError:
                    if errors == "raise":
                        raise ImportError(
                            f"Missing required class: '{class_name}'"
                        ) from None
            except (ImportError, ModuleNotFoundError, ValueError):
                if errors == "ignore":
                    return None
        pypi_name = PYPI_NAMES.get(module_name, None)
        pypi_str = f"(PyPI: '{pypi_name}')" if pypi_name else ""
        if origin_module:
            err = f"Module '{origin_module}' requires '{module_name}' {pypi_str} to be installed."
        else:
            err = f"Missing required module: '{module_name}'"
        if errors == "raise":
            raise ImportError(err) from None
        elif errors == "warn":
            print("Warning:", err)
        return None


err = try_import("google.cloud.logging", "logging", errors="ignore")
try_import("google.cloud.logging.handlers.transports", "logging", errors="ignore")
if err is not None:
    from google.cloud import logging as gcp_logging
    from google.cloud.logging.handlers.transports import SyncTransport

    if os.getenv("PLATFORM", "") in ["GCP", "local"]:
        client = gcp_logging.Client()
        handler = gcp_logging.handlers.CloudLoggingHandler(
            client, name="gcp_pal", transport=SyncTransport
        )
        client.setup_logging()


def log(*args, **kwargs):
    """
    Function for logging to Google Cloud Logs. Logs a message as usual, and logs a dictionary of data as jsonPayload.

    Args:
        *args (list): list of elements to "print" to google cloud logs.
        **kwargs (dict): dictionary of elements to "print" to google cloud logs.

    Examples:
    >>> log("Hello, world!")
    message: "Hello, world!"
    >>> log("Hello, world!", {"a": 1, "b": 2})
    message: "Hello, world!"
    payload: {"a": 1, "b": 2}
    """

    # Use these environment variables as payload to log to Google Cloud Logs
    env_keys = ["PLATFORM"]
    env_data = {key: os.getenv(key, None) for key in env_keys}
    log_data = {k: v for k, v in env_data.items() if v is not None}

    # If any arguments are a dictionary, add it to the log_data so it can be queried in Google Cloud Logs
    for arg in args:
        if isinstance(arg, dict):
            log_data.update(arg)
        log_data["message"] = " ".join([str(a) for a in args])

    if os.getenv("PLATFORM", "Local") in ["GCP"]:
        logging.info(log_data)
    else:
        # If running locally, use a normal print
        print(log_data["message"], **kwargs)


def force_list(x):
    """
    Force x to be a list
    """
    if isinstance(x, LIST_LIKE_TYPES):
        return x
    else:
        return [x]


def get_dict_items(d, item_type="value"):
    """
    Get the items of a dictionary as a list of tuples.

    Args:
    - d (dict): The dictionary.
    - item_type (str): The type of item to get. Can be "key" or "value".

    Returns:
    - list: The items of the dictionary.

    Examples:
    >>> get_dict_items({"a": 1, "b": 2, "c": 3}, item_type="key")
    ["a", "b", "c"]
    >>> get_dict_items({"a": 1, "b": 2, "c": 3}, item_type="value")
    [1, 2, 3]
    >>> get_dict_items({"a": 1, "b": {"c": 2}, "d": 3}, item_type="value")
    [1, 2, 3]
    >>> get_dict_items({"a": 1, "b": {"c": 2}, "d": 3}, item_type="key")
    ["a", "b", "c", "d"]
    >>> get_dict_items({"a": 1, "b": [1, [2, [3, 4]]]}, item_type="value")
    [1, 1, 2, 3, 4]
    >>> get_dict_items({"a": 1, "b": {"c": 2}, "d": [3, {"e": 4}]}, item_type="value")
    [1, 2, 3, 4]
    """
    output = []
    if isinstance(d, dict):
        for key, value in d.items():
            # {key: value}
            if item_type == "key":
                # {<key>: value}
                output.append(key)
                if isinstance(value, (dict, list)):
                    # {key: {<key>: value}}
                    output.extend(get_dict_items(value, item_type))
            elif item_type == "value":
                if isinstance(value, (dict, list)):
                    # {key: {key: <value>}}
                    output.extend(get_dict_items(value, item_type))
                else:
                    # {key: <value>}
                    output.append(value)
    elif isinstance(d, list):
        # [value, value, ...]
        for i in d:
            if isinstance(i, (dict, list)):
                # [{key: value}, {key: value}, ...]
                output.extend(get_dict_items(i, item_type))
            elif item_type == "value":
                # [<value>, <value>, ...]
                output.append(i)
    else:
        # <value>
        output.append(d)
    return output


def is_series(obj):
    """
    Check if an object is a pandas Series without importing pandas.

    Args:
    - obj: The object to check.

    Returns:
    - bool: Whether the object is a pandas Series.
    """
    return type(obj).__name__ == "Series"


def is_dataframe(obj):
    """
    Check if an object is a pandas DataFrame without importing pandas.

    Args:
    - obj: The object to check.

    Returns:
    - bool: Whether the object is a pandas Series.
    """
    return type(obj).__name__ == "DataFrame"


def is_numpy_array(obj):
    """
    Check if an object is a numpy array without importing numpy.

    Args:
    - obj: The object to check.

    Returns:
    - bool: Whether the object is a numpy array.
    """
    return str(type(obj)).startswith("<class 'numpy.")


def is_pyarrow_schema(obj):
    """
    Check if an object is a pyarrow schema without importing pyarrow.

    Args:
    - obj: The object to check.

    Returns:
    - bool: Whether the object is a pandas Series.
    """
    if not obj:
        return False
    if str(type(obj)) == "<class 'pyarrow.lib.Schema'>":
        return True
    elif str(type(obj)) == "<class 'pyarrow.lib.Field'>":
        return True
    elif str(type(obj)) == "<class 'pyarrow.lib.DataType'>":
        return True
    if isinstance(obj, dict):
        return all([is_pyarrow_schema(x) for x in obj.values()])
    elif isinstance(obj, list):
        return all([is_pyarrow_schema(x) for x in obj])
    return False


def is_bigquery_schema(obj):
    """
    Check if an object is a bigquery schema without importing google.cloud.bigquery.

    Args:
    - obj: The object to check.

    Returns:
    - bool: Whether the object is a pandas Series.
    """
    if not obj:
        return False
    if isinstance(obj, list):
        return all([is_bigquery_schema(x) for x in obj])
    return str(type(obj)) == "<class 'google.cloud.bigquery.schema.SchemaField'>"


def is_python_schema(obj):
    """
    Check if an object is a python schema without importing google.cloud.bigquery.

    Args:
    - obj: The object to check.

    Returns:
    - bool: Whether the object is a pandas Series.
    """
    if not obj:
        return False
    if isinstance(obj, dict):
        return all([is_python_schema(x) for x in obj.values()])
    elif isinstance(obj, list):
        return all([is_python_schema(x) for x in obj])
    return str(type(obj)) == "<class 'type'>"


def reverse_dict(d, errors="ignore"):
    """
    Reverse a dictionary. If the dictionary is not one-to-one, the last value in the dictionary will be the one that is kept.

    Args:
    - d (dict): The dictionary to reverse.

    Returns:
    - dict: The reversed dictionary.
    """
    output = {v: k for k, v in d.items()}
    if len(output) < len(d):
        if errors == "ignore":
            print("Warning: Dictionary is not one-to-one.")
        else:
            raise ValueError("Dictionary is not one-to-one.")
    return output


def orient_dict(d, orientation=""):
    """
    Orient a dictionary so that it can be used as a pandas DataFrame.

    Args:
    - d (dict): The dictionary to orient.
    - orientation (str): The orientation to use. Can be "columns" or "index".

    Returns:
    - dict: The oriented dictionary.

    Examples:
    >>> orient_dict({"a": [1, 2], "b": [3, 4]}, orientation="index")
    [{"a": 1, "b": 3}, {"a": 2, "b": 4}]
    >>> orient_dict([{"a": 1, "b": 3}, {"a": 2, "b": 4}], orientation="columns")
    """
    if orientation == "index" and isinstance(d, dict):
        keys = list(d.keys())
        n_values = len(force_list(d[keys[0]]))
        range_n = range(n_values)
        output = [{k: force_list(d[k])[n] for k in d} for n in range_n]
    elif orientation == "columns" and isinstance(d, list):
        keys = list(d[0].keys())
        n_values = len(d)
        range_n = range(n_values)
        output = {k: [d[i][k] for i in range_n] for k in keys}
    else:
        output = d
    return output


def get_auth_default(allow_none=False, errors="raise"):
    """
    Get the default project from google.auth.default() and store it in an environment variable.

    Args:
    - errors (str): The error handling method. Can be "raise" or "warn".

    Returns:
    - str: The default project.
    """
    try_import("google.auth", "get_default_project")
    import google.auth as google_auth

    project = os.getenv("_GOOGLE_AUTH_DEFAULT_PROJECT", None)
    credentials = os.getenv("_GOOGLE_AUTH_DEFAULT_CREDENTIALS", None)
    if project is not None:
        try:
            credentials = google_auth.credentials.Credentials.from_json(credentials)
        except AttributeError:
            pass
        return credentials, project

    credentials, project = google_auth.default()
    if project is None:
        err = "No default project found. Please set the PROJECT environment variable."
        if errors == "raise":
            raise ValueError(err)
        else:
            print("Warning:", err)
        return credentials, project

    os.environ["_GOOGLE_AUTH_DEFAULT_PROJECT"] = project
    try:
        os.environ["_GOOGLE_AUTH_DEFAULT_CREDENTIALS"] = credentials.to_json()
    except AttributeError:
        pass
    print(f"Obtained default project: {project}")
    return credentials, project


def zip_directory(directory, output_file=None, omit_root=True, ignore_file=None):
    """
    Zip a directory.

    Args:
    - directory (str): The directory to zip.
    - output_file (str): The output file.
    """
    from zipfile import ZipFile

    if not output_file:
        output_file = f"{directory}.zip"
    with ZipFile(output_file, "w") as z:
        for root, dirs, files in os.walk(directory):
            for file in files:
                if file == os.path.basename(output_file):
                    # Want to omit the zip file itself to avoid infinite recursion
                    continue
                if omit_root:
                    out_name = os.path.relpath(os.path.join(root, file), directory)
                else:
                    out_name = os.path.join(root, file)
                z.write(os.path.join(root, file), out_name)
    return output_file


def get_all_kwargs(locals_kwargs):
    """
    Get all kwargs from a locals() dictionary.

    Args:
    - locals_kwargs (dict): The dictionary of local variables (locals() output).

    Returns:
    - dict: The dictionary of kwargs.
    """

    all_kwargs = locals_kwargs.copy()
    try:
        all_kwargs.pop("self")
    except KeyError:
        pass
    try:
        kwargs = all_kwargs.pop("kwargs")
    except KeyError:
        kwargs = {}
    all_kwargs = {**all_kwargs, **kwargs}
    return all_kwargs


def load_yaml(file_path):
    """
    Load a YAML file.

    Args:
    - file_path (str): The path to the file.

    Returns:
    - dict: The contents of the file.
    """
    try_import("yaml", "load_yaml")
    import yaml

    with open(file_path, "r") as f:
        return yaml.safe_load(f)


def jprint(x, sort_keys=False, indent=3):
    """
    Pretty print a json object. Basically alias for print(json.dumps(x, indent=3))
    """
    try:
        log(json.dumps(x, indent=indent, sort_keys=sort_keys))
    except Exception as e:
        log(x)


def get_default_arg(arg_name):
    """
    Get a default argument from the environment variables.
    """
    from gcp_pal.config import DEFAULT_ARGS

    arg_dict = {
        "project": os.environ.get("GCP_PAL_PROJECT") or get_auth_default()[1],
        "location": os.environ.get("GCP_PAL_LOCATION")
        or DEFAULT_ARGS.get("location", None),
    }
    output = arg_dict.get(arg_name, None)
    return output


class JSON:
    """
    Class for operating JSON files
    """

    def __init__(self, path, platform=os, input_dict=None):
        self.path = path
        if isinstance(path, dict):
            input_dict = path
        self.platform = platform
        self.open = self.platform.open if platform != os else open
        self.exists = self.platform.exists if platform != os else os.path.exists
        self.input_dict = input_dict

    def load(self, allow_empty=True):
        """
        Load a json file as a dict
        """
        log("Loading", self.path)
        if allow_empty and not self.exists(self.path):
            return {}
        with self.open(self.path, "r") as f:
            return json.load(f)

    def write(self, data, **kwargs):
        """
        Write a json file
        """
        kwargs.setdefault("indent", 3)
        kwargs.setdefault("sort_keys", True)
        self.platform.makedirs(os.path.dirname(self.path), exist_ok=True)
        with self.open(self.path, mode="w") as f:
            json.dump(data, f, **kwargs)

    def print(self, sort_keys=False, indent=3):
        """
        Print the contents of a json file
        """
        if isinstance(self.path, dict):
            jprint(self.path, sort_keys=sort_keys, indent=indent)


class YAML:
    """
    Class for operating YAML files
    """

    def __init__(self, file_name, platform=os):
        from gcp_pal.utils import ModuleHandler

        self.file_name = file_name
        self.yaml = ModuleHandler("yaml").please_import(who_is_calling="YAML")

        self.path = file_name
        self.platform = platform
        self.open = self.platform.open if platform != os else open
        self.exists = self.platform.exists if platform != os else os.path.exists

    def load(self, allow_empty=True):
        if allow_empty and not self.exists(self.path):
            return {}
        with open(self.file_name, "r") as file:
            return self.yaml.safe_load(file)

    def write(self, data, sort_keys=True):
        with open(self.file_name, "w") as file:
            self.yaml.dump(data, file, default_flow_style=False, sort_keys=sort_keys)

    def print(self):
        if isinstance(self.file_name, dict):
            jprint(self.file_name, sort_keys=False, indent=3)
        else:
            with open(self.file_name, "r") as file:
                jprint(self.yaml.safe_load(file), sort_keys=False, indent=3)


def load_env_vars_file(env_vars_file, allow_empty=True):
    lower_file = env_vars_file.lower()
    if lower_file.endswith(".yaml") or lower_file.endswith(".yml"):
        try:
            loaded_vars = YAML(env_vars_file).load(allow_empty=allow_empty)
        except:
            loaded_vars = {}
    elif lower_file.endswith(".json"):
        try:
            loaded_vars = JSON(env_vars_file).load(allow_empty=allow_empty)
        except:
            loaded_vars = {}
    if allow_empty and not loaded_vars:
        loaded_vars = {}
    return loaded_vars


if __name__ == "__main__":
    d = {"a": 1, "b": {"c": 2}, "d": [3, {"e": 4}]}
    JSON("red/test.json").write(d)



import os
from gcp_pal.utils import try_import

from typing import List
from urllib.parse import unquote

from gcp_pal.utils import is_dataframe, force_list, ModuleHandler


class Parquet:
    """
    A class for working with Parquet files in Google Cloud Storage.
    """

    def __init__(self, path: str, **kwargs):
        self.path = path
        self.pa = ModuleHandler("pyarrow").please_import(who_is_calling="Parquet")
        self.pq = ModuleHandler("pyarrow.parquet").please_import(
            who_is_calling="Parquet"
        )

    def write(
        self,
        data,
        partition_cols: List[str] = None,
        schema=None,
        basename_template="{i}.parquet",
        **kwargs,
    ):
        """
        Write a DataFrame to a Parquet file in Google Cloud Storage.

        Args:
        - data: The data to write.
        - partition_cols: The columns to partition by.
        - schema: The schema to write.
        - basename_template: The template for the Parquet file names.
        - kwargs: Additional arguments to pass to `pyarrow` or `pandas`.
        """
        if schema is None:
            from gcp_pal.schema import Schema

            schema = Schema(data).pyarrow()
        if partition_cols:
            partition_cols = force_list(partition_cols)
            if set(partition_cols) == set(data.columns):
                msg = (
                    f"Partition columns {partition_cols} are the same as data columns."
                )
                msg += "\nThis might cause unexpected behaviour with BigQuery external tables."
                # msg += "\nAdding an `__index` column to the data."
                # log(msg)
                # data["__index"] = range(len(data))
                # if schema:
                #     schema = schema.append(pa.field("__index", pa.int64()))
            return self._write_partitioned(
                data=data,
                partition_cols=partition_cols,
                schema=schema,
                basename_template=basename_template,
                **kwargs,
            )
        else:
            return self._write_single(data, schema=schema, **kwargs)

    def _write_partitioned(
        self, data, partition_cols, schema, basename_template, **kwargs
    ):
        """
        Write a DataFrame to a partitioned Parquet file in Google Cloud Storage.

        Args:
        - data: The data to write.
        - partition_cols: The columns to partition by.
        - schema: The schema to write.
        - basename_template: The template for the Parquet file names.
        - kwargs: Additional arguments to pass to `pyarrow.parquet.write_to_dataset`.
        """
        if is_dataframe(data):
            self.pq.write_to_dataset(
                table=self.pa.Table.from_pandas(data),
                root_path=self.path,
                partition_cols=partition_cols,
                schema=schema,
                basename_template=basename_template,
                **kwargs,
            )
        else:
            raise ValueError(f"Data of type {type(data)} is not supported yet.")

    def _write_single(self, data, schema, **kwargs):
        """
        Write a DataFrame to a single Parquet file in Google Cloud Storage.

        Args:
        - data: The data to write.
        - schema: The schema to write.
        - kwargs: Additional arguments to pass to `pyarrow.parquet.write_table`.
        """
        if is_dataframe(data):
            return data.to_parquet(
                self.path,
                index=False,
                engine="pyarrow",
                **kwargs,
            )
        else:
            raise ValueError(f"Data of type {type(data)} is not supported yet.")

    def read(
        self,
        allow_empty=True,
        schema=None,
        filters=None,
        columns=None,
        read_partitions_only=False,
        **kwargs,
    ):
        """
        Read a Parquet file from Google Cloud Storage.

        Args:
        - allow_empty: Whether to allow empty data.
        - schema: The schema to read.
        - filters: The filters to apply.
        - use_bigquery: Whether to use BigQuery.
        - columns: The columns to read.
        - read_partitions_only: If true, it will read the data from partition names only using `glob`.
        """
        if read_partitions_only:
            from gcp_pal.schema import Schema

            df = _get_partitions_df(self.path)
            schema = Schema(schema).pandas()
            df = df.astype(schema)
        else:
            if isinstance(schema, dict):
                from gcp_pal.schema import Schema

                schema = Schema(schema).pyarrow()
            try:
                df = self.pq.read_table(
                    self.path,
                    columns=columns,
                    filters=filters,
                    schema=schema,
                    **kwargs,
                ).to_pandas()
            except FileNotFoundError:
                try_import("pandas")
                import pandas as pd

                df = pd.DataFrame()
        if df.empty and not allow_empty:
            raise ValueError(
                f"File {self.path} is empty or not found, and allow_empty is False"
            )
        if "__index" in df.columns:
            df = df.drop(columns="__index")
        df = self.fix_column_types(df, filters)

        return df

    def generate_filters(self, filters: dict = None):
        """
        Generate the filters

        Args:
        - filters: The filters to apply. If the filters are a dictionary, it will be converted to a list of tuples.

        Returns:
        - list[tuple]

        Examples:
        >>> Parquet().generate_filters({"a": [1, 2], "b": [3, 4], "c": 5})
        [('a', 'in', [1, 2]), ('b', 'in', [3, 4]), ('c', 'in', [5])]
        """
        if filters is None:
            return None
        if isinstance(filters, list):
            return filters

        file_filters = []
        for column, value in filters.items():
            value = force_list(value)
            file_filters.append((column, "in", value))
        return file_filters

    def fix_column_types(self, df, filters, replace_underscore=True):
        """
        Ensure nothing strange happens with the column types of the df.

        Args:
        - df: The DataFrame to fix.
        - filters: The filters to apply.
        - replace_underscore: Whether to replace underscores with spaces in string columns.

        Returns:
        - pd.DataFrame

        Examples:
        >>> Parquet().fix_column_types(df, filters={"a": [1, 2], "b": [3, 4], "c": 5})
        """
        if (not filters) or df.empty:
            return df
        for c, c_type in [(x[0], type(x[1])) for x in filters]:
            df[c] = df[c].astype(c_type)
            if replace_underscore and (c_type == str):
                df[c] = df[c].str.replace("_", " ")
        return df


def _get_partitions_df(path):
    """
    Get the partitions of the parquet file
    """
    try_import("pandas", "parquet._get_partitions_df")
    import pandas as pd

    all_partition_paths = _get_all_partition_paths(path)
    all_partitions = [x.split("/") for x in all_partition_paths]
    all_partitions = [dict(y.split("=") for y in x if "=" in y) for x in all_partitions]
    df = pd.DataFrame(all_partitions)
    df = df.map(unquote)
    return df


def _get_all_partition_paths(path):
    """
    Get the partitions of the parquet file
    """
    from gcp_pal import Storage

    partition_cols = _get_partition_cols(path)
    glob_query = os.path.join(path, *["*"] * len(partition_cols))
    all_partitions = Storage().glob(glob_query)
    return all_partitions


def _get_partition_cols(path):
    """
    Get the partitions of the parquet file without reading the files
    """
    from gcp_pal import Storage

    glob_query = os.path.join(path, "*")
    all_paths = Storage().glob(glob_query)
    while all_paths:
        glob_query = all_paths[0] + "/*"
        all_paths = Storage().glob(glob_query)
    glob_query = glob_query.split("/")
    partition_path = [x.split("=")[0] for x in glob_query if "=" in x]
    return partition_path


if __name__ == "__main__":
    import pandas as pd
    from gcp_pal import Storage
    from gcp_pal.schema import Schema

    success = {}
    data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    schema = Schema(data).pyarrow()
    bucket_name = f"test_bucket_0b759c89-d519-4f14-971e-cc09727e5266"
    file_name = f"gs://{bucket_name}/file.parquet"
    # Storage().create_bucket(bucket_name)
    # Parquet(file_name).write(data, partition_cols=["a", "b"])

    read_df = Parquet(file_name).read(read_partitions_only=True, schema=schema)
    print(read_df)
    print(data)
    print(read_df.dtypes)
    print(data.dtypes)



import json

from gcp_pal.utils import (
    log,
    ClientHandler,
    ModuleHandler,
    get_default_arg,
    get_all_kwargs,
)


class PubSub:
    """
    Class for pushing a payload to a topic

    Examples:
    - `PubSub("topic").publish("data")` -> Publish `"data"` to "topic"
    - `PubSub("topic").publish({"key": "value"})` -> Publish `{"key": "value"}` to "topic"
    """

    def __init__(
        self,
        path: str = "",
        topic: str = None,
        subscription: str = None,
        project=None,
    ):
        """
        Args:
        - `path` (str): Path to the topic or project. Default is "". Supported formats:
            - "projects/my-project/topics/my-topic/subscription/my-subscription"
            - "projects/my-project/topics/my-topic"
            - "projects/my-project"
            - "my-project/my-topic/my-subscription"
            - "my-project/my-topic"
            - "my-project"
        - `topic` (str): Name of the topic
        - `subscription` (str): Name of the subscription
        - `project` (str): Project ID
        """
        self.subscription = None
        self.topic_id = None
        self.project = None
        self.path = path
        if path.startswith("projects/"):
            path = path.split("/")[1::2]
            path = "/".join(path)
            self.path = path
        try:
            self.project, self.topic_id, self.subscription = path.split("/")
        except ValueError:
            pass
        try:
            self.project, self.topic_id = path.split("/")
        except ValueError:
            pass
        if isinstance(path, str) and path.count("/") == 0:
            self.project = path
        self.topic_id = self.topic_id or topic
        self.subscription = self.subscription or subscription
        self.project = self.project or project or get_default_arg("project")
        self.level = self._set_level()
        self.path = self._set_path()
        self.parent = f"projects/{self.project}"
        if self.level == "topic":
            self.parent = f"{self.parent}/topics/{self.topic_id}"

        self.pubsub = ModuleHandler("google.cloud").please_import(
            "pubsub_v1", who_is_calling="PubSub"
        )
        self.publisher = ClientHandler(self.pubsub.PublisherClient).get()
        self.subscriber = ClientHandler(self.pubsub.SubscriberClient).get()
        self.types = self.pubsub.types
        self.exceptions = ModuleHandler("google.api_core").please_import(
            "exceptions", who_is_calling="PubSub"
        )
        self.topic_path = self.publisher.topic_path(self.project, self.topic_id)
        self.duration = ModuleHandler("google.protobuf").please_import(
            "duration_pb2", who_is_calling="PubSub"
        )

    def __repr__(self):
        return f"PubSub({self.topic_id})"

    def _set_level(self):
        if self.subscription and self.project:
            return "subscription"
        elif self.topic_id and self.project:
            return "topic"
        elif self.project:
            return "project"
        else:
            raise ValueError("Invalid level.")

    def _set_path(self):
        path = self.project
        if self.topic_id:
            path = f"{path}/{self.topic_id}"
        if self.subscription:
            path = f"{path}/{self.subscription}"
        return path

    def ls_topics(self, full_name=False, include_project=False):
        """
        Lists all topics in a project.

        Args:
        - `full_name` (bool): If False, returns names in format "{project}/{topic}".
                              Otherwise, returns full names "projects/{project}/topics/{topic}".
                              Default is False.
        - `include_project` (bool): If False, returns names in format "{topic}".
        """
        topics = self.publisher.list_topics(request={"project": self.parent})
        output = [topic.name for topic in topics]
        if not full_name:
            output = ["/".join(topic.split("/")[1::2]) for topic in output]
            if not include_project:
                output = ["/".join(topic.split("/")[1:]) for topic in output]
        return output

    def ls_subscriptions(self, full_name=False, include_project=False):
        """
        Lists all subscriptions for a topic.

        Args:
        - `full_name` (bool): If False, returns names in format "{project}/{topic}/{subscription}".
                              Otherwise, returns full names "projects/{project}/topics/{topic}/subscriptions/{subscription}".
                              Default is False.
        - `include_project` (bool): If False, returns names in format "{topic}/{subscription}".
        """
        if self.level == "topic":
            subscriptions = self.publisher.list_topic_subscriptions(
                request={"topic": self.parent}
            )
        elif self.level == "project":
            subscriptions = self.subscriber.list_subscriptions(
                request={"project": self.parent}
            )
            subscriptions = [subscription.name for subscription in subscriptions]
        output = [subscription for subscription in subscriptions]
        if not full_name:
            output = [
                "/".join(subscription.split("/")[1::2]) for subscription in output
            ]
            if not include_project:
                output = [
                    "/".join(subscription.split("/")[1:]) for subscription in output
                ]
        return output

    def ls(self, full_name=False, include_project=False):
        """
        Lists all topics or subscriptions.

        Args:
        - `full_name` (bool): If False, returns names in format "{project}/{topic}" or "{topic}/{subscription}".
                              Otherwise, returns full names "projects/{project}/topics/{topic}" or "projects/{project}/topics/{topic}/subscriptions/{subscription}".
                              Default is False.
        - `include_project` (bool): If False, returns names in format "{topic}" or "{subscription}".
        """
        if self.level == "topic":
            return self.ls_subscriptions(
                full_name=full_name, include_project=include_project
            )
        elif self.level == "project":
            return self.ls_topics(full_name=full_name, include_project=include_project)
        else:
            raise ValueError(f"Invalid level: {self.level}")

    def publish(self, data) -> None:
        if isinstance(data, dict):
            data = json.dumps(data)
        try:
            publish_future = self.publisher.publish(self.parent, data.encode("utf-8"))
            result = publish_future.result()
            log(
                f"PubSub - Published message: {result} to {self.topic_id} in {self.project}."
            )
            return result
        except Exception as e:
            log(f"PubSub - An error occurred: {e}")
            return

    def create_topic(
        self,
        topic=None,
        labels=None,
        schema=None,
        encoding=None,
        message_retention_duration=None,
        if_exists="ignore",
    ):
        """
        Create a topic.

        Args:
        - `topic` (str): Name of the topic. Can be specified in the constructor.
        - `labels` (dict): Labels for the topic. Default is None.
        - `schema` (str): Name of the schema against which to validate message payloads. Has the form `projects/{project}/schemas/{schema}`.
        - `encoding` (str): The encoding of the messages validated against the schema. Default is None.
        - `message_retention_duration` (int): The duration (in seconds) for which the topic retains unacknowledged messages. Default is None.
        """
        topic = topic or self.topic_id
        schema_settings = (
            self.types.SchemaSettings(schema=schema, encoding=encoding)
            if schema
            else None
        )
        if message_retention_duration:
            message_retention_duration = self.duration.Duration(
                seconds=message_retention_duration
            )
        log(f"PubSub - Creating topic: {self.topic_id} in {self.project}...")
        topic_resource = self.types.Topic(
            name=self.parent,
            schema_settings=schema_settings,
            labels=labels,
        )
        try:
            result = self.publisher.create_topic(request=topic_resource)
        except self.exceptions.AlreadyExists as e:
            if if_exists != "ignore":
                raise e
            log(f"PubSub - Topic already exists: {self.topic_id} in {self.project}")
            result = self.publisher.get_topic(request={"topic": self.parent})
            return result
        log(f"PubSub - Topic created: {result.name} in {self.project}")
        return result

    def create_subscription(
        self,
        subscription=None,
        topic=None,
        ack_deadline_seconds=None,
        retain_acked_messages=None,
        message_retention_duration=None,
        labels=None,
        minimum_backoff=10,
        maximum_backoff=600,
        detached=False,
        filter=None,
        table_id=None,
        if_exists="ignore",
        **kwargs,
    ):
        """
        Create a subscription.

        Args:
        - `subscription` (str): Name of the subscription. Can be specified in the constructor.
        - `topic` (str): Name of the topic. Can be specified in the constructor.
        - `ack_deadline_seconds` (int): The maximum time after receiving a message that the subscriber must acknowledge the message. Default is None.
        - `retain_acked_messages` (bool): Whether to retain acknowledged messages. Default is None.
        - `message_retention_duration` (int): How long to retain unacknowledged messages. Default is None.
        - `labels` (dict): Labels for the subscription. Default is None.
        - `minimum_backoff` (int): Minimum delay between redelivery attempts. Default is 10.
        - `maximum_backoff` (int): Maximum delay between redelivery attempts. Default is 600.
        - `detached` (bool): If True, the subscription will be detached from the topic. Default is False.
        - `filter` (str): A filter expression that determines which messages are delivered to the subscription. Default is None.
        - `table_id` (str): The BigQuery table ID to which the subscription's data will be written. Default is None.
        - `if_exists` (str): If "ignore", does not raise an error if the subscription already exists. Default is "ignore".
        - `kwargs`: Additional arguments to be passed to the relevant configurations.

        Returns:
        - The reference to the created subscription.
        """
        log(f"PubSub - Creating subscription: {self.subscription} in {self.project}...")
        subscription = subscription or self.subscription
        topic = topic or self.topic_id
        name = f"projects/{self.project}/subscriptions/{subscription}"
        retry_policy = self.types.RetryPolicy(
            minimum_backoff=self.duration.Duration(seconds=minimum_backoff),
            maximum_backoff=self.duration.Duration(seconds=maximum_backoff),
        )
        if message_retention_duration:
            duration_pb2 = ModuleHandler("google.protobuf").please_import(
                "duration_pb2", who_is_calling="PubSub"
            )
            message_retention_duration = duration_pb2.Duration(
                seconds=message_retention_duration
            )
        table = table_id
        all_kwargs = get_all_kwargs(locals())
        bigquery_config, storage_config, push_config = self._split_subscription_kwargs(
            all_kwargs
        )
        bigquery_config = (
            self.types.BigQueryConfig(**bigquery_config) if bigquery_config else None
        )
        storage_config = (
            self.types.CloudStorageConfig(**storage_config) if storage_config else None
        )
        push_config = self.types.PushConfig(**push_config) if push_config else None
        subscription_resource = self.types.Subscription(
            name=name,
            topic=f"projects/{self.project}/topics/{topic}",
            ack_deadline_seconds=ack_deadline_seconds,
            retain_acked_messages=retain_acked_messages,
            message_retention_duration=message_retention_duration,
            labels=labels,
            retry_policy=retry_policy,
            detached=detached,
            filter=filter,
            bigquery_config=bigquery_config,
            cloud_storage_config=storage_config,
            push_config=push_config,
        )
        try:
            result = self.subscriber.create_subscription(request=subscription_resource)
        except self.exceptions.AlreadyExists as e:
            if if_exists != "ignore":
                raise e
            log(
                f"PubSub - Subscription already exists: {self.subscription} in {self.project}"
            )
            result = self.subscriber.get_subscription(request={"subscription": name})
            return result
        log(f"PubSub - Subscription created: {result.name} in {self.project}.")
        return subscription

    def _split_subscription_kwargs(self, kwargs):
        """
        Splits the subscription kwargs into function, service and build kwargs.

        Returns:
        - (dict, dict, dict) The function, service and build kwargs.
        """
        bigquery_kwargs = {}
        cloud_storage_kwargs = {}
        push_kwargs = {}
        for key, value in kwargs.items():
            if value is None:
                continue
            if key in self.types.BigQueryConfig.__annotations__:
                bigquery_kwargs[key] = value
            if key in self.types.PushConfig.__annotations__:
                push_kwargs[key] = value
            if key in self.types.CloudStorageConfig.__annotations__:
                cloud_storage_kwargs[key] = value
        return bigquery_kwargs, cloud_storage_kwargs, push_kwargs

    def delete_topic(self, errors="ignore"):
        """
        Delete a topic.

        Args:
        - `errors` (str): If "ignore", does not raise an error if the topic does not exist. Default is "ignore".

        Returns:
        - True if the topic no longer exists.
        """
        log(f"PubSub - Deleting topic: {self.topic_id} in {self.project}...")
        try:
            self.publisher.delete_topic(request={"topic": self.topic_path})
        except self.exceptions.NotFound as e:
            if errors != "ignore":
                raise e
            log(
                f"PubSub - Topic not found to delete: {self.topic_id} in {self.project}"
            )
            return
        log(f"PubSub - Topic deleted: {self.topic_id} in {self.project}")
        return True

    def delete_subscription(self, errors="ignore"):
        """
        Delete a subscription.

        Args:
        - `errors` (str): If "ignore", does not raise an error if the subscription does not exist. Default is "ignore".

        Returns:
        - True if the subscription no longer exists.
        """
        parent = f"projects/{self.project}/subscriptions/{self.subscription}"
        log(f"PubSub - Deleting subscription: {self.subscription} in {self.project}...")
        try:
            self.subscriber.delete_subscription(request={"subscription": parent})
        except self.exceptions.NotFound as e:
            if errors != "ignore":
                raise e
            log(
                f"PubSub - Subscription not found to delete: {self.subscription} in {self.project}"
            )
            return
        log(f"PubSub - Subscription deleted: {self.subscription} in {self.project}")
        return True

    def delete(self, errors="ignore"):
        """
        Delete a topic or subscription.

        Args:
        - `errors` (str): If "ignore", does not raise an error if the topic or subscription does not exist. Default is "ignore".
        """
        if self.level == "topic":
            return self.delete_topic(errors=errors)
        elif self.level == "subscription":
            return self.delete_subscription(errors=errors)
        else:
            raise ValueError(f"Invalid level: {self.level}")

    def create(self, **kwargs):
        """
        Create a topic or subscription.

        Args:
        - `kwargs`: Keyword arguments to be passed to the relevant `create()` method.

        Returns:
        - The topic or subscription.
        """
        if self.level == "topic":
            return self.create_topic(**kwargs)
        elif self.level == "subscription":
            return self.create_subscription(**kwargs)
        else:
            raise ValueError(f"Invalid level: {self.level}")

    def get(self, level=None):
        """
        Get a reference to the topic or subscription.

        Args:
        - `level` (str): Level to get. Either "topic" or "subscription". Default is `self.level`.

        Returns:
        - The topic or subscription.
        """
        level = level or self.level
        if level == "topic":
            return self.publisher.get_topic(request={"topic": self.parent})
        elif level == "subscription":
            parent = f"projects/{self.project}/subscriptions/{self.subscription}"
            return self.subscriber.get_subscription(request={"subscription": parent})
        else:
            raise ValueError(f"Invalid level: {level}")

    def exists(self):
        try:
            self.get()
            return True
        except self.exceptions.NotFound:
            return False


if __name__ == "__main__":
    PubSub().ls()
    PubSub().ls_subscriptions()
    # PubSub("test_topic").publish("data")
    # PubSub("test_topic").publish({"key": "value"})



from gcp_pal.bigquery import BigQuery
from gcp_pal.cloudfunctions import CloudFunctions
from gcp_pal.cloudrun import CloudRun
from gcp_pal.cloudscheduler import CloudScheduler
from gcp_pal.firestore import Firestore
from gcp_pal.pubsub import PubSub
from gcp_pal.pylogging import Logging
from gcp_pal.request import Request
from gcp_pal.schema import Schema
from gcp_pal.storage import Storage
from gcp_pal.storage import Parquet
from gcp_pal.secretmanager import SecretManager
from gcp_pal.project import Project
from gcp_pal.dataplex import Dataplex
from gcp_pal.datastore import Datastore
from gcp_pal.artifactregistry import ArtifactRegistry



from gcp_pal.utils import ModuleHandler, ClientHandler, get_default_arg, log


class Datastore:

    _clients = {}

    def __init__(
        self,
        namespace: str = None,
        kind: str = None,
        entity: dict = None,
        database: str = None,
        project: str = None,
        location: str = "europe-west2",
    ):
        """
        Datastore class to interact with Google Cloud Datastore.

        Args:
        - database (str): Database of the entity. Default is None.
        - namespace (str): Namespace of the entity. Default is None.
        - kind (str): Kind of the entity. Default is None.
        - entity (dict): Entity to be stored. Default is None.
        - project (str): Project of the entity. Default is "PROJECT" environment variable or the gcloud default project.
        - location (str): Location of the entity. Default is "europe-west2".
        """
        self.project = project or get_default_arg("project")
        self.database = database
        self.namespace = namespace
        self.kind = kind
        self.entity = entity
        self.location = location

        self._set_up_attributes()

        self.datastore = ModuleHandler("google.cloud").please_import(
            "datastore", who_is_calling="Datastore"
        )

        self.client = ClientHandler(self.datastore.Client).get(
            project=self.project,
            namespace=self.namespace,
            database=self.database,
        )
        self.admin = ModuleHandler("google.cloud").please_import(
            "firestore_admin_v1", who_is_calling="Firestore"
        )
        self.admin_client = ClientHandler(self.admin.FirestoreAdminClient).get()

    def _set_up_attributes(self):
        """
        Sets up the attributes of the Datastore object. Extracts the project, database, namespace, kind and entity from the path.
        """
        self.level = self._get_level()
        self.path = self._get_path()

    def _get_level(self):
        """
        Returns the level of the path. It can be one of: project, database, namespace, kind, entity.

        Returns:
        - str: Level of the path.
        """
        if self.entity:
            if not self.namespace:
                self.namespace = "default"
            if not self.database:
                self.database = "default"
            return "entity"
        elif self.kind:
            if not self.namespace:
                self.namespace = "default"
            if not self.database:
                self.database = "default"
            return "kind"
        elif self.namespace:
            if not self.database:
                self.database = "default"
            return "namespace"
        elif self.database:
            return "database"
        else:
            return "project"

    def _get_path(self):
        """
        Returns the path of the entity in the format: "project.database.namespace/kind/entity".

        Returns:
        - str: Path of the entity.
        """
        if self.entity:
            return f"{self.project}.{self.database}/{self.namespace}/{self.kind}/{self.entity}"
        elif self.kind:
            return f"{self.project}.{self.database}/{self.namespace}/{self.kind}"
        elif self.namespace:
            return f"{self.project}.{self.database}/{self.namespace}"
        elif self.database:
            return f"{self.project}.{self.database}"
        else:
            return f"{self.project}"

    def __repr__(self):
        paths = {
            "project": self.project,
            "database": self.database,
            "namespace": self.namespace,
            "kind": self.kind,
            "entity": self.entity,
        }
        paths = [f"{key}={value}" for key, value in paths.items() if value]
        output = f"Datastore({', '.join(paths)})"
        return output

    def query(self, filters: dict = None, order: str = None, limit: int = None):
        """
        Returns a query object to interact with the Datastore.

        Returns:
        - Query: Query object to interact with the Datastore.
        """
        query = self.client.query(kind=self.kind, namespace=self.namespace)

        if filters:
            for key, value in filters.items():
                query.add_filter(key, "=", value)

        if order:
            query.order = order

        if limit:
            query.limit = limit

        return query

    def fetch(
        self,
        filters: dict = None,
        order: str = None,
        limit: int = None,
        as_dict: bool = False,
    ):
        """
        Fetches the entities from the Datastore.

        Args:
        - filters (dict): Filters to be applied to the query. Default is None.
        - order (str): Order of the entities. Default is None.
        - limit (int): Limit of the entities. Default is None.

        Returns:
        - list: List of entities fetched from the Datastore.
        """
        if not self.namespace:
            self.namespace = "default"
        if not self.database:
            self.database = "default"
        log(f"Datastore - Fetching {self}...")

        query = self.query(filters=filters, order=order, limit=limit)
        if as_dict:
            entities = [dict(entity) for entity in query.fetch()]
        else:
            entities = list(query.fetch())
        return entities

    def fetch_one(self, filters: dict = None, order: str = None, as_dict: bool = False):
        """
        Fetches one entity from the Datastore.

        Args:
        - filters (dict): Filters to be applied to the query. Default is None.
        - order (str): Order of the entities. Default is None.

        Returns:
        - Entity: Entity fetched from the Datastore.
        """
        entities = self.fetch(filters=filters, order=order, limit=1, as_dict=as_dict)
        if entities:
            return entities[0]
        return None

    def read(self, **kwargs):
        """
        Alias for fetch method.

        Args:
        - kwargs: Keyword arguments to be passed to the fetch method.

        Returns:
        - Entity: Entity fetched from the Datastore.
        """
        return self.fetch(**kwargs)

    def ls_databases(self, full_path=False):
        """
        List all databases in a Firestore project.
        """
        databases = self.admin_client.list_databases(parent=f"projects/{self.project}")
        output = [database.name for database in databases.databases]
        if not full_path:
            output = [database.split("/")[-1] for database in output]
        log(f"Firestore - databases listed.")
        return output


if __name__ == "__main__":
    ds = Datastore(database="database1")
    print(ds)
    breakpoint()
